@@ -263,7 +263,7 @@ sub pre_modify
                         }
                     }
                     foreach my $rwattr (keys %replaced) {
-                        push(@mod_list, ($action, $rwattr, @{$replaced{$rwattr}}));
+                        push(@mod_list, ($action, $rwattr, $self->_unique(@{$replaced{$rwattr}})));
                     }
                 } else {
                     push(@mod_list, ($action, $attr));
@@ -316,7 +316,7 @@ sub pre_add
                     return -1;
                 }
             }
-            ${$entryStrp}[0] = join("\n", @line)."\n";
+            ${$entryStrp}[0] = join("\n", $self->_unique(@line))."\n";
         }
     }
 
@@ -465,7 +465,7 @@ sub _rewriteParse
 
         # do functions
         my @substs = ($substline);
-        my @rwmaps = ($substline =~ /%{([^(]*\([^)]*\).*)}/g);
+        my @rwmaps = ($substline =~ /%{([^(]*\(.*\))}/gs);
         foreach my $rwmap (@rwmaps) {
             my @values;
             my $key = lc($rwmap);
