@@ -201,7 +201,7 @@ sub search
     }
 
     # get entry of data container
-    if ($base =~ /^$self->{suffix}$/) {
+    if ($base =~ /^$self->{suffix}$/i) {
         if ($scope != 1) {
             my $entry = $self->{contentrystr};
             if ($self->parseFilter($filter, $entry)) {
@@ -213,28 +213,35 @@ sub search
 
     foreach my $objinfo (@objs) {
         my ($obj, $pkeys) = @{$objinfo};
-        my $objbase = $base =~ /^$self->{suffix}$/ ? $obj->{suffix} : $base;
+        my $objbase = $base =~ /^$self->{suffix}$/i ? $obj->{suffix} : $base;
         my $objscope;
         my $entry;
         my $key;
 
-        if ($scope == 1 && $base =~ /^$self->{suffix}$/ && $obj->{entrystr}) {
+        if ($scope == 1 && $base =~ /^$self->{suffix}$/i && $obj->{entrystr}) {
             $objscope = 0;
         } else {
             $objscope = $scope;
         }
 
         # search base entry
-        ($rc, $key, $entry) = $self->_baseSearch($obj, $pkeys, $objbase, $objscope, $deref, $sizeLim, $timeLim, $filter);
-        if ($entry) {
-            push(@match_entries, $entry);
-            $sizeLim = $sizeLim - $entry;
-        }
-
+        ($rc, $key, $entry) = $self->_baseSearch($obj, $pkeys, $objbase, 0, $deref, $sizeLim, $timeLim, undef);
         if ($rc) {
             last;
         }
 
+        # don't search children if entry doesn't exist
+        if (!$entry && $base !~ /^$self->{suffix}$/i) {
+            if ($scope == 0) {
+                return LDAP_NO_SUCH_OBJECT;
+            } else {
+                next;
+            }
+        } elsif ($objscope !=1 && $self->parseFilter($filter, $entry)) {
+            push(@match_entries, $entry);
+            $sizeLim = $sizeLim - $entry;
+        }
+
         # search entries below base
         my @entries;
         my $keys;
@@ -618,6 +625,10 @@ sub parseFilter
     my $self = shift;
     my ($filter, $entry) = @_;
 
+    if (!$entry) {
+        return 0;
+    }
+
     if (!$filter) {
         return 1;
     }
@@ -642,12 +653,13 @@ sub parseFilter
                 next;
             }
 
+            my $assval =  decode('utf8', $args->{assertionValue});
             if ($op eq 'equalityMatch') {
-                $rc = ($val =~ /^$args->{assertionValue}$/i);
+                $rc = ($val =~ /^$assval$/i);
             } elsif ($op eq 'greaterOrEqual') {
-                $rc = ($val ge $args->{assertionValue});
+                $rc = ($val ge $assval);
             } elsif ($op eq 'lessOrEqual') {
-                $rc = ($val le $args->{assertionValue});
+                $rc = ($val le $assval);
             }
 
             if ($rc) {
@@ -657,11 +669,14 @@ sub parseFilter
         return $rc;
     } elsif ($op eq 'substrings') {
         if (defined($args->{substrings}[0]{initial})) {
-            return $entry =~ /^$args->{type}: $args->{substrings}[0]{initial}.*$/mi;
+            my $substr = decode('utf8', $args->{substrings}[0]{initial});
+            return $entry =~ /^$args->{type}: $substr.*$/mi;
         } elsif (defined($args->{substrings}[0]{final})) {
-            return $entry =~ /^$args->{type}: .*$args->{substrings}[0]{final}$/mi;
+            my $substr = decode('utf8', $args->{substrings}[0]{final});
+            return $entry =~ /^$args->{type}: .*$substr$/mi;
         } else {
-            return $entry =~ /^$args->{type}: .*$args->{substrings}[0]{any}.*$/mi;
+            my $substr = decode('utf8', $args->{substrings}[0]{any});
+            return $entry =~ /^$args->{type}: .*$substr.*$/mi;
         }
     } elsif ($op eq 'present') {
         return $entry =~ /^$args: /mi;
@@ -702,6 +717,11 @@ sub buildEntryStr
     return $entry;
 }
 
+sub manageDIT
+{
+    return 0;
+}
+
 
 sub _checkConfig
 {
@@ -844,6 +864,21 @@ sub _checkConfig
                     }
                 }
             }
+
+            # check paramset
+            foreach my $func (keys %{$oconf->{plugin}}) {
+                my $plugin = $oconf->{plugin}{$func};
+                foreach my $key1 (keys %{$plugin->{param}}) {
+                    if (defined($plugin->{param}{$key1}->{type}) &&
+                        $plugin->{param}{$key1}->{type} eq 'paramset' &&
+                            defined($conf->{paramset}{$plugin->{param}{$key1}->{value}})) {
+                        my $paramset = $conf->{paramset}{$plugin->{param}{$key1}->{value}};
+                        foreach my $key2 (keys %{$paramset->{param}}) {
+                            $plugin->{param}{$key2} = $paramset->{param}{$key2};
+                        }
+                    }
+                }
+            }
         }
 
         # check container's link
@@ -1008,7 +1043,7 @@ sub _searchObjects
         return (LDAP_NO_SUCH_OBJECT, @objs);
     }
 
-    if ($base =~ /^$self->{suffix}$/) {
+    if ($base =~ /^$self->{suffix}$/i) {
         if ($scope != 0) {
             foreach my $oname (keys %{$self->{object}}) {
                 if (defined($self->{object}{$oname}->{suffix})) {
@@ -1268,6 +1303,59 @@ sub _funcParse
     return $str;
 }
 
+sub _doPlugin
+{
+    my $self = shift;
+    my ($op, $obj, @args) = @_;
+    my $oconf = $obj->{conf};
+    my @match_keys;
+    my @match_entries;
+    my $rc = LDAP_SUCCESS;
+
+    foreach my $func (keys %{$oconf->{plugin}}) {
+        my $plugin = $oconf->{plugin}{$func};
+        my $error;
+
+        if (!defined($plugin->{op}) || $plugin->{op}[0] eq $op) {
+            if ($op eq 'search') {
+                my $keys;
+                my @entries;
+
+                eval "(\$rc, \$error, \$keys, \@entries) = LISM::Storage::$func(\$self, \$oconf, \$plugin->{param}, \@args)";
+                for (my $i = 0; $i < @entries; $i++) {
+                    if ($entries[$i] !~ /^dn: .*,$args[1]\n/i) {
+                        next;
+                    }
+
+                    my $level = split(/,/, ($entries[$i] =~ /^dn: (.*),?$self->{suffix}\n/i)[0]);
+                    if ($obj->{level} == $level &&
+                        $self->parseFilter($args[3], $entries[$i])) {
+                        push(@match_keys, ${$keys}[$i]);
+                        push(@match_entries, $entries[$i]);
+                    }
+                }
+            } else {
+                eval "(\$rc, \$error) = LISM::Storage::$func(\$self, \$oconf, \$plugin->{param}, \@args)";
+            }
+            if ($@) {
+                $self->log(level => 'err', message => "plugin $func failed in $op operation: $@");
+                $rc = LDAP_OTHER;
+                last;
+            } elsif ($rc) {
+                $self->log(level => 'err', message => "plugin $func returns error in $op operation: $error");
+                $rc = LDAP_OTHER;
+                last;
+            }
+        }
+    }
+
+    if ($op eq 'search') {
+        return ($rc, \@match_keys, @match_entries);
+    } else {
+        return $rc;
+    }
+}
+
 sub _pwdFormat
 {
     my $self = shift;
