@@ -374,6 +374,11 @@ sub hashPasswd
     return $hashpw;
 }
 
+sub manageDIT
+{
+    return 1;
+}
+
 
 sub _getConnect
 {
@@ -463,7 +468,7 @@ sub _do_search
     }
 
     # get entry of data container
-    if ($base =~ /^$self->{suffix}$/) {
+    if ($base =~ /^$self->{suffix}$/i) {
         if ($scope != 1) {
             my $entry = $self->{contentrystr};
             if ($self->parseFilter($filter, $entry)) {
