@@ -215,21 +215,6 @@ sub _checkConfig
                 }
             }
         }
-
-        # check paramset
-        foreach my $func (keys %{$oconf->{plugin}}) {
-            my $plugin = $oconf->{plugin}{$func};
-            foreach my $key1 (keys %{$plugin->{param}}) {
-                if (defined($plugin->{param}{$key1}->{type}) &&
-                    $plugin->{param}{$key1}->{type} eq 'paramset' &&
-                        defined($conf->{paramset}{$plugin->{param}{$key1}->{value}})) {
-                    my $paramset = $conf->{paramset}{$plugin->{param}{$key1}->{value}};
-                    foreach my $key2 (keys %{$paramset->{param}}) {
-                        $plugin->{param}{$key2} = $paramset->{param}{$key2};
-                    }
-                }
-            }
-        }
     }
 
     return 0;
@@ -428,21 +413,9 @@ sub _objModify
         }
 
         if (defined($oconf->{plugin})) {
-            foreach my $func (keys %{$oconf->{plugin}}) {
-                my $plugin = $oconf->{plugin}{$func};
-                my $error;
-                if (!defined($plugin->{op}) || $plugin->{op}[0] eq 'modify') {
-                    eval "(\$rc, \$error) = LISM::Storage::$func(\$self, \$oconf, \$plugin->{param}, \$pkeys, \$key, \$dn, \@list)";
-                    if ($@) {
-                        $self->log(level => 'err', message => "plugin $func failed in modify operation: $@");
-                        $rc = LDAP_OTHER;
-                        last DO;
-                    } elsif ($rc) {
-                        $self->log(level => 'err', message => "plugin $func returns error in modify operation: $error");
-                        $rc = LDAP_OTHER;
-                        last DO;
-                    }
-                }
+            $rc = $self->_doPlugin('modify', $obj, $pkeys, $key, $dn, @list);
+            if ($rc) {
+                last DO;
             }
         }
 
@@ -521,14 +494,12 @@ sub _objModify
                 }
 
                 if (defined($oconf->{attr}{$attr}->{addproc})) {
-                    for (my $i = 0; $i < @values; $i++) {
-                        if ($self->_addAttrValues($oconf, $key, $pkey, $attr, $values[$i], $dn)) {
-                            $rc = LDAP_OPERATIONS_ERROR;
-                            last DO;
-                        }
+                    if ($self->_addAttrValues($oconf->{attr}{$attr}, $key, $pkey, $dn, '', @values)) {
+                        $rc = LDAP_OPERATIONS_ERROR;
+                        last DO;
                     }
                 }
-            } elsif( $action eq "DELETE" ) {
+            } elsif($action eq "DELETE") {
                 # check whether the value exists
                 for (my $i = 0; $i < @values; $i++) {
                     my $value = $values[$i];
@@ -564,11 +535,9 @@ sub _objModify
                         }
                     }
 
-                    foreach my $value (@values) {
-                        if ($self->_delAttrValues($oconf, $key, $pkey, $attr, $value)) {
-                            $rc = LDAP_OPERATIONS_ERROR;
-                            last DO;
-                        }
+                    if ($self->_delAttrValues($oconf->{attr}{$attr}, $key, $pkey, '', @values)) {
+                        $rc = LDAP_OPERATIONS_ERROR;
+                        last DO;
                     }
                 }
             } elsif( $action eq "REPLACE" ) {
@@ -593,47 +562,42 @@ sub _objModify
                 }
 
                 if (defined($oconf->{attr}{$attr}->{addproc}) || defined($oconf->{attr}{$attr}->{delproc})) {
-                    my @old_vals = ($entry =~ /^$attr: (.*)$/mgi);
-                    my @existing_vals;
+                    my @old_vals = ($entry =~ /^$attr: (.*)$/gmi);
+                    my @add_vals;
+                    my @delete_vals;
 
                     # delete the old values which exists
-                    foreach my $old_val (@old_vals) {
-                        my $match = 0;
-                        for (my $i = 0; $i < @values; $i++) {
-                            my $value = $values[$i];
-                            $value =~ s/([.*+?\[\]()|\^\$\\])/\\$1/g;
-                            if ($old_val =~ /^$value$/i) {
-                                $match = 1;
-                                next;
+                    foreach my $value (@values) {
+                        my $tmpval = $value;
+                        $tmpval =~ s/([.*+?\[\]()|\^\$\\])/\\$1/g;
+
+                        my $valmatch = 0;
+                        my $i = 0;
+                        for ($i = 0; $i < @old_vals; $i++) {
+                            if ($old_vals[$i] =~ /^$tmpval$/i) {
+                                $valmatch = 1;
+                                last;
                             }
                         }
-                        if ($match) {
-                            push(@existing_vals, $old_val);
-                            next;
-                        }
-
-                        # escape special character
-                        $old_val =~ s/'/''/g;
-                        $old_val =~ s/\\/\\\\/g;
-
-                        if ($self->_delAttrValues($oconf, $key, $pkey, $attr, $old_val)) {
-                            $rc = LDAP_OPERATIONS_ERROR;
-                            last DO;
+                        if ($valmatch) {
+                            splice(@old_vals, $i, 1);
+                        } else {
+                            # escape special character
+                            $value =~ s/'/''/g;
+                            $value =~ s/\\/\\\\/g;
+                            push(@add_vals, $value);
                         }
                     }
+                    @delete_vals = @old_vals;
 
-                    # add the new values doesn't exist
-                    for (my $i = 0; $i < @values; $i++) {
-                        my $value = $values[$i];
-                        $value =~ s/([.*+?\[\]()|\^\$\\])/\\$1/g;
-                        if (grep(/^$value$/i, @existing_vals)) {
-                            next;
-                        }
+                    if ($self->_addAttrValues($oconf->{attr}{$attr}, $key, $pkey, $dn, '', @add_vals)) {
+                        $rc = LDAP_OPERATIONS_ERROR;
+                        last DO;
+                    }
 
-                        if ($self->_addAttrValues($oconf, $key, $pkey, $attr, $values[$i], $dn)) {
-                            $rc = LDAP_OPERATIONS_ERROR;
-                            last DO;
-                        }
+                    if ($self->_delAttrValues($oconf->{attr}{$attr}, $key, $pkey, '', @delete_vals)) {
+                        $rc = LDAP_OPERATIONS_ERROR;
+                        last DO;
                     }
                 }
             }
@@ -676,19 +640,9 @@ sub _objAdd
     }
 
     if (defined($oconf->{plugin})) {
-        foreach my $func (keys %{$oconf->{plugin}}) {
-            my $plugin = $oconf->{plugin}{$func};
-            my $error;
-            if (!defined($plugin->{op}) || $plugin->{op}[0] eq 'add') {
-                eval "(\$rc, \$error) = LISM::Storage::$func(\$self, \$oconf, \$plugin->{param}, \$pkeys, \$dn, \$entryStr)";
-                if ($@) {
-                    $self->log(level => 'err', message => "plugin $func failed in add operation: $@");
-                    return LDAP_OTHER;
-                } elsif ($rc) {
-                    $self->log(level => 'err', message => "plugin $func returns error in add operation: $error");
-                    return LDAP_OTHER;
-                }
-            }
+        $rc = $self->_doPlugin('add', $obj, $pkeys, $dn, $entryStr);
+        if ($rc) {
+            return $rc;
         }
     }
 
@@ -810,12 +764,10 @@ sub _objAdd
         if (!defined($oconf->{noop}) || !grep(/^add$/i, @{$oconf->{noop}})) {
             # add the values in the attribute's table
             foreach my $attr (keys %attrs) {
-                foreach my $value (@{$attrs{$attr}}) {
-                    if ($rc = $self->_addAttrValues($oconf, $key, $pkey, $attr, $value, $dn, $entryStr)) {
-                        $self->log(level => 'err', message => "Adding values of $attr failed");
-                        $rc = LDAP_OPERATIONS_ERROR;
-                        last DO;
-                    }
+                if ($self->_addAttrValues($oconf->{attr}{$attr}, $key, $pkey, $dn, $entryStr, @{$attrs{$attr}})) {
+                    $self->log(level => 'err', message => "Adding values of $attr failed");
+                    $rc = LDAP_OPERATIONS_ERROR;
+                    last DO;
                 }
             }
         }
@@ -887,7 +839,7 @@ sub _objDelete
             # delete the values from the attribute's table
             foreach my $attr (keys %{$oconf->{attr}}) {
                 if (defined($oconf->{attr}{$attr}) && defined($oconf->{attr}{$attr}->{delproc})) {
-                    if ($rc = $self->_delAttrValues($oconf, $key, $pkey, $attr, '')) {
+                    if ($self->_delAttrValues($oconf->{attr}{$attr}, $key, $pkey, '', '')) {
                         $rc = LDAP_OPERATIONS_ERROR;
                         last DO;
                     }
@@ -948,21 +900,9 @@ sub _objDelete
         }
 
         if (defined($oconf->{plugin})) {
-            foreach my $func (keys %{$oconf->{plugin}}) {
-                my $plugin = $oconf->{plugin}{$func};
-                my $error;
-                if (!defined($plugin->{op}) || $plugin->{op}[0] eq 'delete') {
-                    eval "(\$rc, \$error) = LISM::Storage::$func(\$self, \$oconf, \$plugin->{param}, \$pkeys, \$key, \$dn)";
-                    if ($@) {
-                        $self->log(level => 'err', message => "plugin $func failed in delete operation: $@");
-                        $rc = LDAP_OTHER;
-                        last DO;
-                    } elsif ($rc) {
-                        $self->log(level => 'err', message => "plugin $func returns error in delete operation: $error");
-                        $rc = LDAP_OTHER;
-                        last DO;
-                    }
-                }
+            $rc = $self->_doPlugin('delete', $obj, $pkeys, $key, $dn);
+            if ($rc) {
+                last DO;
             }
         }
     }
@@ -1137,91 +1077,19 @@ sub _getAttrValues
 sub _addAttrValues
 {
     my $self = shift;
-    my ($oconf, $key, $pkey, $attr, $value, $dn, $entryStr) = @_;
+    my ($aconf, $key, $pkey, $dn, $entryStr, @values) = @_;
     my $conf = $self->{_config};
 
-    if (!defined($oconf->{attr}{$attr}->{addproc})) {
+    if (!defined($aconf->{addproc})) {
         return -1;
     }
 
-    foreach my $addproc (@{$oconf->{attr}{$attr}->{addproc}}) {
-        my $sql = $addproc;
-        $sql =~ s/\%o/$key/g;
-        $sql =~ s/\%p/$pkey/g;
-        if (defined($oconf->{attr}{$attr}->{oname})) {
-            my ($rc, $aobj, $attrpkeys) = $self->_getObject($value);
-            if ($rc) {
-                return -1;
-            }
-
-            my $attrkey;
-            ($rc, $attrkey) =$self->_baseSearch($aobj, $attrpkeys, $value, 0, 0, 1, 0, undef);
-            if ($rc) {
-                $self->log(level => 'err', message => "Can't get id of $value in the table");
-                return -1;
-            } elsif (!$attrkey) {
-		$self->log(level => 'err', message => "Id of $value doesn't exist in the table");
-                return 1;
-            }
-            $sql =~ s/\%a/$attrkey/g;
-
-            if (defined($aobj->{conf}->{parentrel}) &&
-                $aobj->{conf}->{parentrel}[0] eq 'tight') {
-                my $contdn;
-                ($contdn = $value) =~ s/^[^,]+,//;
-                my ($rc, $cobj, $contpkeys) = $self->_getObject($contdn);
-                if ($rc) {
-                    return -1;
-                }
-
-                my $contkey;
-                ($rc, $contkey) =$self->_baseSearch($cobj, $contpkeys, $contdn, 0, 0, 1, 0, undef);
-                if ($rc) {
-                    $self->log(level => 'err', message => "Can't get id of $contdn in the table");
-                    return -1;
-                } elsif (!$contkey) {
-                    $self->log(level => 'err', message => "Id of $contdn doesn't exist in the table");
-                    return 1;
-                }
-                $sql =~ s/\%c/$contkey/g;
-            }
-        } else {
-            $sql =~ s/\%a/$value/g;
-        }
-
-        $sql = $self->_funcParse($sql, $dn, $entryStr);
-
-        # multibyte value
-        $sql = encode($conf->{mbcode}[0], $sql);
-
-        # add the values to the attribute's table
-	my $sth = $self->{db}->prepare($sql);
-        if (!$sth->execute) {
-            $self->log(level => 'err', message => "Adding $attr values by \"$sql\" failed: ".$sth->errstr);
-            return -1;
-        }
-        $sth->finish;
-    }
-
-    return 0;
-}
-
-sub _delAttrValues
-{
-    my $self = shift;
-    my ($oconf, $key, $pkey, $attr, $value, $dn) = @_;
-    my $conf = $self->{_config};
-
-    if (!defined($oconf->{attr}{$attr}->{delproc})) {
-        return 1;
-    }
-
-    foreach my $delproc (@{$oconf->{attr}{$attr}->{delproc}}) {
-        my $sql = $delproc;
-        $sql =~ s/\%o/$key/g;
-        $sql =~ s/\%p/$pkey/g;
-        if ($value) {
-            if (defined($oconf->{attr}{$attr}->{oname})) {
+    foreach my $value (@values) {
+        foreach my $addproc (@{$aconf->{addproc}}) {
+            my $sql = $addproc;
+            $sql =~ s/\%o/$key/g;
+            $sql =~ s/\%p/$pkey/g;
+            if (defined($aconf->{oname})) {
                 my ($rc, $aobj, $attrpkeys) = $self->_getObject($value);
                 if ($rc) {
                     return -1;
@@ -1230,7 +1098,7 @@ sub _delAttrValues
                 my $attrkey;
                 ($rc, $attrkey) =$self->_baseSearch($aobj, $attrpkeys, $value, 0, 0, 1, 0, undef);
                 if ($rc) {
-		    $self->log(level => 'err', message => "Can't get id of $value from the table");
+                    $self->log(level => 'err', message => "Can't get id of $value in the table");
                     return -1;
                 } elsif (!$attrkey) {
 		    $self->log(level => 'err', message => "Id of $value doesn't exist in the table");
@@ -1261,23 +1129,99 @@ sub _delAttrValues
             } else {
                 $sql =~ s/\%a/$value/g;
             }
-        } else {
-            $sql =~ s/\w*\s*=\s*'?\%a'?/1 = 1/g;
-            $sql =~ s/\w*\s*=\s*'?\%c'?/1 = 1/g;
+
+            $sql = $self->_funcParse($sql, $dn, $entryStr);
+
+            # multibyte value
+            $sql = encode($conf->{mbcode}[0], $sql);
+
+            # add the values to the attribute's table
+	    my $sth = $self->{db}->prepare($sql);
+            if (!$sth->execute) {
+                $self->log(level => 'err', message => "Adding values by \"$sql\" failed: ".$sth->errstr);
+                return -1;
+            }
+            $sth->finish;
         }
+    }
 
-        $sql = $self->_funcParse($sql, $dn);
+    return 0;
+}
 
-        # multibyte value
-        $sql = encode($conf->{mbcode}[0], $sql);
+sub _delAttrValues
+{
+    my $self = shift;
+    my ($aconf, $key, $pkey, $dn, @values) = @_;
+    my $conf = $self->{_config};
 
-        # delete the values from the attribute's table
-        my $sth = $self->{db}->prepare($sql);
-        if (!$sth->execute) {
-            $self->log(level => 'err', message => "Deleting $attr values by \"$sql\" failed: ".$sth->errstr);
-            return -1;
+    if (!defined($aconf->{delproc})) {
+        return 1;
+    }
+
+    foreach my $value (@values) {
+        foreach my $delproc (@{$aconf->{delproc}}) {
+            my $sql = $delproc;
+            $sql =~ s/\%o/$key/g;
+            $sql =~ s/\%p/$pkey/g;
+            if ($value) {
+                if (defined($aconf->{oname})) {
+                    my ($rc, $aobj, $attrpkeys) = $self->_getObject($value);
+                    if ($rc) {
+                        return -1;
+                    }
+
+                    my $attrkey;
+                    ($rc, $attrkey) =$self->_baseSearch($aobj, $attrpkeys, $value, 0, 0, 1, 0, undef);
+                    if ($rc) {
+		        $self->log(level => 'err', message => "Can't get id of $value from the table");
+                        return -1;
+                    } elsif (!$attrkey) {
+		        $self->log(level => 'err', message => "Id of $value doesn't exist in the table");
+                        return 1;
+                    }
+                    $sql =~ s/\%a/$attrkey/g;
+
+                    if (defined($aobj->{conf}->{parentrel}) &&
+                        $aobj->{conf}->{parentrel}[0] eq 'tight') {
+                        my $contdn;
+                        ($contdn = $value) =~ s/^[^,]+,//;
+                        my ($rc, $cobj, $contpkeys) = $self->_getObject($contdn);
+                        if ($rc) {
+                            return -1;
+                        }
+
+                        my $contkey;
+                        ($rc, $contkey) =$self->_baseSearch($cobj, $contpkeys, $contdn, 0, 0, 1, 0, undef);
+                        if ($rc) {
+                            $self->log(level => 'err', message => "Can't get id of $contdn in the table");
+                            return -1;
+                        } elsif (!$contkey) {
+                            $self->log(level => 'err', message => "Id of $contdn doesn't exist in the table");
+                            return 1;
+                        }
+                        $sql =~ s/\%c/$contkey/g;
+                    }
+                } else {
+                    $sql =~ s/\%a/$value/g;
+                }
+            } else {
+                $sql =~ s/\w*\s*=\s*'?\%a'?/1 = 1/g;
+                $sql =~ s/\w*\s*=\s*'?\%c'?/1 = 1/g;
+            }
+
+            $sql = $self->_funcParse($sql, $dn);
+
+            # multibyte value
+            $sql = encode($conf->{mbcode}[0], $sql);
+
+            # delete the values from the attribute's table
+            my $sth = $self->{db}->prepare($sql);
+            if (!$sth->execute) {
+                $self->log(level => 'err', message => "Deleting values by \"$sql\" failed: ".$sth->errstr);
+                return -1;
+            }
+            $sth->finish;
         }
-        $sth->finish;
     }
 
     return 0;
